#lang racket
(require rackunit)
(require "extras.rkt")
(require 2htdp/universe)   
(require 2htdp/image)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; These images are from
;;;
;;;     https://en.wikipedia.org/wiki/File:Walk-cycle-poses.jpg
;;;
;;; and licensed under the Creative Commons Attribution 3.0 license.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define RIGHT-FORWARD (bitmap "Walk1.png"))
(define RIGHT-STANDING (bitmap "Walk2.png"))
(define RIGHT-BACK (bitmap "Walk3.png"))
(define RIGHT-PASSING (bitmap "Walk4.png"))

(define RIGHT-LIST
  (list RIGHT-FORWARD RIGHT-STANDING RIGHT-BACK RIGHT-PASSING))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (main rate)
  (big-bang
   (initial-world)
   (on-tick world-after-tick rate)
   (on-key world-after-key-event)
   (on-mouse world-after-mouse-event)
   (to-draw world-to-scene)))

(define X-GUY 200)
(define Y-GUY 300)

(define CANVAS-HEIGHT 600)
(define CANVAS-WIDTH 800)

(define (initial-world)
  (new WorldState%
       [widgets
        (list (new Guy% [x X-GUY] [y Y-GUY] [phase RIGHT-FORWARD]))]))

(define (world-after-tick world)
  (send world after-tick))

(define (world-after-key-event world ke)
  (send world after-key-event ke))

(define (world-after-mouse-event world mx my me)
  (send world after-mouse-event mx my me))

(define (world-to-scene world)
  (send world to-scene))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interfaces from Lesson 9.2
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define Widget<%>
  (interface ()
    
    ;; -> Widget<%>
    ;; GIVEN: no arguments
    ;; RETURNS: the state of this object that should follow after a tick
    
    after-tick

    ;; Integer Integer -> Widget<%>
    ;; GIVEN: x and y coordinates for a location
    ;; RETURNS: the state of this object taht should follow the specified
    ;;     mouse event at the given location

    after-button-down
    after-button-up
    after-drag

    ;; KeyEvent -> Widget<%>
    ;; GIVEN: a key event
    ;; RETURNS: the state of this object that should follow after the
    ;;     given key event

    after-key-event

    ;; Scene -> Scene
    ;; GIVEN: a scene
    ;; RETURNS: a scene like the given one, but with this object painted on it

    add-to-scene
    ))

(define WorldState<%>
  (interface ()

    ;; -> WorldState<%>
    ;; GIVEN: no arguments
    ;; RETURNS: the state of the world at the next tick

    after-tick

    ;; Integer Integer MouseEvent -> WorldState<%>
    ;; GIVEN: a location and what kind of mouse event
    ;; RETURNS: the state of the world that should follow the given mouse
    ;;     event at the given location

    after-mouse-event

    ;; KeyEvent -> WorldState<%>
    ;; GIVEN: a key event
    ;; RETURNS: the state of the world that should follow the given key event

    after-key-event

    ;; -> Scene
    ;; GIVEN: no arguments
    ;; RETURNS: a scene that depicts the world

    to-scene
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; New stuff.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define WorldState%
  (class* object% (WorldState<%>)
    
    (init-field widgets) ;; widgets is a ListOfWidget<%>
    (super-new)

    ;; -> WorldState<%>
    ;; GIVEN: no arguments
    ;; RETURNS: the state of the world at the next tick

    (define/public (after-tick)
      (new WorldState%
           [widgets
            (map (lambda (widget)
                   (send widget after-tick))
                 widgets)]))

    ;; Integer Integer MouseEvent -> WorldState<%>
    ;; GIVEN: a location and what kind of mouse event
    ;; RETURNS: the state of the world that should follow the given mouse
    ;;     event at the given location

    (define/public (after-mouse-event mx my me)
      (new WorldState%
           [widgets
            (map (lambda (widget)
                   (cond ((string=? me "button-down")
                          (send widget button-down mx my))
                         ((string=? me "button-up")
                          (send widget button-up mx my))
                         ((string=? me "drag")
                          (send widget drag mx my))
                         (else
                          widget)))
                 widgets)]))

    ;; KeyEvent -> WorldState<%>
    ;; GIVEN: a key event
    ;; RETURNS: the state of the world that should follow the given key event

    (define/public (after-key-event ke)
      (new WorldState%
           [widgets
            (map (lambda (widget)
                   (send widget after-key-event ke))
                 widgets)]))

    ;; -> Scene
    ;; GIVEN: no arguments
    ;; RETURNS: a scene that depicts the world

    (define/public (to-scene)
      (foldr
       (lambda (widget scene)
         (send widget add-to-scene scene))
       (empty-scene CANVAS-WIDTH CANVAS-HEIGHT)
       widgets))

    ))

(define Guy%
  (class* object% (Widget<%>)

    (init-field x y phase) ;; coordinates

    (super-new)
    
    ;; -> Widget<%>
    ;; GIVEN: no arguments
    ;; RETURNS: the state of this object that should follow after a tick
    
    (define/public (after-tick)
      (new Guy%
           [x (+ x X-OFFSET)]
           [y (+ y Y-OFFSET)]
           [phase (next-phase phase)]))

    (define X-OFFSET 10)
    (define Y-OFFSET 0)

    ;; Integer Integer -> Widget<%>
    ;; GIVEN: x and y coordinates for a location
    ;; RETURNS: the state of this object taht should follow the specified
    ;;     mouse event at the given location

    (define/public (after-button-down mx my)
      this)

    (define/public (after-button-up mx my)
      this)

    (define/public (after-drag mx my)
      this)

    ;; KeyEvent -> Widget<%>
    ;; GIVEN: a key event
    ;; RETURNS: the state of this object that should follow after the
    ;;     given key event

    (define/public (after-key-event ke)
      this)

    ;; Scene -> Scene
    ;; GIVEN: a scene
    ;; RETURNS: a scene like the given one, but with this object painted on it

    (define/public (add-to-scene scene)
      (place-image phase
                   x y scene))
       
    ))

(define (next-phase phase)
  (cond ((equal? phase RIGHT-FORWARD)
         RIGHT-STANDING)
        ((equal? phase RIGHT-STANDING)
         RIGHT-BACK)
        ((equal? phase RIGHT-BACK)
         RIGHT-PASSING)
        ((equal? phase RIGHT-PASSING)
         RIGHT-FORWARD)))
